@@ -69,6 +69,13 @@ namespace {
                                     : (*(s)++ + 9))))
 #endif
 
+#if PERL_VERSION < 14
+    inline void GPD_warn_sv(pTHX_ SV *mess) {
+        warn("%" SVf, mess);
+    }
+    #define warn_sv(sv) GPD_warn_sv(aTHX_ (sv))
+#endif
+
     inline void set_bool(pTHX_ SV *target, bool value) {
         if (value)
             sv_setiv(target, 1);
@@ -305,6 +312,8 @@ bool Mapper::DecoderHandlers::on_end_sub_message(DecoderHandlers *cxt, const int
 }
 
 bool Mapper::DecoderHandlers::on_end_map_entry(DecoderHandlers *cxt, const int *field_index) {
+    THX_DECLARE_AND_GET;
+
     size_t size = cxt->items.size();
     HV *hash = (HV *) cxt->items[size - 3];
     SV *key = (SV *) cxt->items[size - 2];
@@ -499,6 +508,7 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
     encode_defaults = message_def->syntax() == UPB_SYNTAX_PROTO2 &&
         options.encode_defaults;
     check_enum_values = options.check_enum_values;
+    warn_context = WarnContext::get(aTHX);
 
     if (!decoder_handlers->SetEndMessageHandler(UpbMakeHandler(DecoderHandlers::on_end_message)))
         croak("Unable to set upb end message handler for %s", message_def->full_name());
@@ -790,6 +800,8 @@ SV *Mapper::encode(SV *ref) {
         croak("It looks like resolve_references() was not called (and please use map() anyway)");
     status.Clear();
     output_buffer.clear();
+    warn_context->clear();
+    warn_context->localize_warning_handler(aTHX);
     SV *result = NULL;
     if (encode(pb_encoder->input(), &status, ref))
         result = newSVpvn(output_buffer.data(), output_buffer.size());
@@ -803,6 +815,8 @@ SV *Mapper::encode_json(SV *ref) {
         croak("It looks like resolve_references() was not called (and please use map() anyway)");
     status.Clear();
     output_buffer.clear();
+    warn_context->clear();
+    warn_context->localize_warning_handler(aTHX);
     SV *result = NULL;
     if (encode(json_encoder->input(), &status, ref))
         result = newSVpvn(output_buffer.data(), output_buffer.size());
@@ -912,7 +926,7 @@ namespace {
             return SvIV(src);
     }
 
-    IV key_iv(const char *key, I32 keylen) {
+    IV key_iv(pTHX_ const char *key, I32 keylen) {
         UV value;
         int numtype = grok_number(key, keylen, &value);
 
@@ -929,7 +943,7 @@ namespace {
         return 0;
     }
 
-    UV key_uv(const char *key, I32 keylen) {
+    UV key_uv(pTHX_ const char *key, I32 keylen) {
         UV value;
         int numtype = grok_number(key, keylen, &value);
 
@@ -1052,13 +1066,14 @@ bool Mapper::encode_from_array(Sink *sink, Status *status, const Mapper::Field &
     G getter;
     S setter(status);
     Sink sub;
-    bool packed = upb_fielddef_isprimitive(fd.field_def) && fd.field_def->packed();
 
     if (!sink->StartSequence(fd.selector.seq_start, &sub))
         return false;
     int size = av_top_index(source) + 1;
 
+    WarnContext::Item &warn_cxt = warn_context->push_level(WarnContext::Array);
     for (int i = 0; i < size; ++i) {
+        warn_cxt.index = i;
         SV **item = av_fetch(source, i, 0);
         if (!item)
             return false;
@@ -1066,6 +1081,7 @@ bool Mapper::encode_from_array(Sink *sink, Status *status, const Mapper::Field &
         if (!setter(aTHX_ &sub, fd, getter(aTHX_ *item)))
             return false;
     }
+    warn_context->pop_level();
 
     return sink->EndSequence(fd.selector.seq_end);
 }
@@ -1077,7 +1093,9 @@ bool Mapper::encode_from_message_array(Sink *sink, Status *status, const Mapper:
     if (!sink->StartSequence(fd.selector.seq_start, &sub))
         return false;
 
+    WarnContext::Item &warn_cxt = warn_context->push_level(WarnContext::Array);
     for (int i = 0; i < size; ++i) {
+        warn_cxt.index = i;
         SV **item = av_fetch(source, i, 0);
         if (!item)
             return false;
@@ -1091,6 +1109,7 @@ bool Mapper::encode_from_message_array(Sink *sink, Status *status, const Mapper:
         if (!sub.EndSubMessage(fd.selector.msg_end))
             return false;
     }
+    warn_context->pop_level();
 
     return sink->EndSequence(fd.selector.seq_end);
 }
@@ -1105,6 +1124,7 @@ namespace {
 }
 
 bool Mapper::encode(Sink *sink, Status *status, SV *ref) const {
+    SvGETMAGIC(ref);
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
         croak("Not an hash reference when encoding a %s value", message_def->full_name());
     HV *hv = (HV *) SvRV(ref);
@@ -1115,7 +1135,9 @@ bool Mapper::encode(Sink *sink, Status *status, SV *ref) const {
     bool tied = SvTIED_mg((SV *) hv, PERL_MAGIC_tied);
     bool ok = true;
     int last_oneof = -1;
+    WarnContext::Item &warn_cxt = warn_context->push_level(WarnContext::Message);
     for (vector<Field>::const_iterator it = fields.begin(), en = fields.end(); it != en; ++it) {
+        warn_cxt.field = &*it;
         HE *he = tied ? hv_fetch_ent_tied(aTHX_ hv, it->name, 0, it->name_hash) :
                         hv_fetch_ent(hv, it->name, 0, it->name_hash);
 
@@ -1143,6 +1165,7 @@ bool Mapper::encode(Sink *sink, Status *status, SV *ref) const {
         else
             ok = ok && encode_nodefaults(sink, status, *it, HeVAL(he));
     }
+    warn_context->pop_level();
 
     if (!sink->EndMessage(status))
         return false;
@@ -1302,13 +1325,13 @@ bool Mapper::encode_key(Sink *sink, Status *status, const Field &fd, const char
         return sink->EndString(fd.selector.str_end);
     }
     case UPB_TYPE_INT32:
-        return sink->PutInt32(fd.selector.primitive, key_iv(key, keylen));
+        return sink->PutInt32(fd.selector.primitive, key_iv(aTHX_ key, keylen));
     case UPB_TYPE_UINT32:
-        return sink->PutUInt32(fd.selector.primitive, key_uv(key, keylen));
+        return sink->PutUInt32(fd.selector.primitive, key_uv(aTHX_ key, keylen));
     case UPB_TYPE_INT64:
-        return sink->PutInt64(fd.selector.primitive, key_iv(key, keylen));
+        return sink->PutInt64(fd.selector.primitive, key_iv(aTHX_ key, keylen));
     case UPB_TYPE_UINT64:
-        return sink->PutInt64(fd.selector.primitive, key_uv(key, keylen));
+        return sink->PutInt64(fd.selector.primitive, key_uv(aTHX_ key, keylen));
     default:
         return false; // just in case
     }
@@ -1334,6 +1357,7 @@ bool Mapper::encode_hash_kv(Sink *sink, Status *status, const char *key, STRLEN
 }
 
 bool Mapper::encode_from_perl_hash(Sink *sink, Status *status, const Field &fd, SV *ref) const {
+    SvGETMAGIC(ref);
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
         croak("Not an hash reference when encoding field '%s'", fd.full_name().c_str());
     HV *hash = (HV *) SvRV(ref);
@@ -1343,6 +1367,7 @@ bool Mapper::encode_from_perl_hash(Sink *sink, Status *status, const Field &fd,
         return false;
 
     hv_iterinit(hash);
+    WarnContext::Item &warn_cxt = warn_context->push_level(WarnContext::Hash);
     while (HE *entry = hv_iternext(hash)) {
         Sink key_value;
         SV *value = HeVAL(entry);
@@ -1363,6 +1388,8 @@ bool Mapper::encode_from_perl_hash(Sink *sink, Status *status, const Field &fd,
             }
         }
 
+        warn_cxt.key = key;
+        warn_cxt.keylen = keylen;
         if (!repeated.StartSubMessage(fd.selector.msg_start, &key_value))
             return false;
         if (!fd.mapper->encode_hash_kv(&key_value, status, key, keylen, value))
@@ -1370,11 +1397,13 @@ bool Mapper::encode_from_perl_hash(Sink *sink, Status *status, const Field &fd,
         if (!repeated.EndSubMessage(fd.selector.msg_end))
             return false;
     }
+    warn_context->pop_level();
 
     return sink->EndSequence(fd.selector.seq_end);
 }
 
 bool Mapper::encode_from_perl_array(Sink *sink, Status *status, const Field &fd, SV *ref) const {
+    SvGETMAGIC(ref);
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVAV)
         croak("Not an array reference when encoding field '%s'", fd.full_name().c_str());
     AV *array = (AV *) SvRV(ref);
@@ -1455,6 +1484,7 @@ bool Mapper::check_from_enum_array(Status *status, const Mapper::Field &fd, AV *
 }
 
 bool Mapper::check(Status *status, SV *ref) const {
+    SvGETMAGIC(ref);
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
         croak("Not an hash reference when checking a %s value", message_def->full_name());
     HV *hv = (HV *) SvRV(ref);
@@ -1515,6 +1545,7 @@ bool Mapper::check(Status *status, const Field &fd, SV *ref) const {
 }
 
 bool Mapper::check_from_perl_array(Status *status, const Field &fd, SV *ref) const {
+    SvGETMAGIC(ref);
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVAV)
         croak("Not an array reference when encoding field '%s'", fd.full_name().c_str());
     AV *array = (AV *) SvRV(ref);
@@ -2002,3 +2033,61 @@ SV *EnumMapper::enum_descriptor() const {
 
     return ref;
 }
+
+WarnContext::WarnContext(pTHX) : chained_handler(NULL) {
+    CV *handler = get_cv("Google::ProtocolBuffers::Dynamic::Mapper::handle_warning", 0);
+
+    warn_handler = (SV*) handler;
+}
+
+void WarnContext::setup(pTHX) {
+    CV *handler = get_cv("Google::ProtocolBuffers::Dynamic::Mapper::handle_warning", 0);
+
+    CvXSUBANY(handler).any_ptr = new WarnContext(aTHX);
+}
+
+WarnContext *WarnContext::get(pTHX) {
+    CV *handler = get_cv("Google::ProtocolBuffers::Dynamic::Mapper::handle_warning", 0);
+
+    return (WarnContext *) CvXSUBANY(handler).any_ptr;
+}
+
+void WarnContext::localize_warning_handler(pTHX) {
+    chained_handler = PL_warnhook;
+    SAVEGENERICSV(PL_warnhook);
+    PL_warnhook = SvREFCNT_inc_simple_NN(warn_handler);
+}
+
+void WarnContext::warn_with_context(pTHX_ SV *warning) const {
+    SV *cxt = sv_2mortal(newSVpvs("While encoding field '"));
+
+    for (Levels::const_iterator it = levels.begin(), en = levels.end(); it != en; ++it) {
+        switch (it->kind) {
+        case Array:
+            sv_catpvf(cxt, "[%d].", it->index);
+            break;
+        case Hash:
+            sv_catpvf(cxt, "{%" SVf "}.", it->index);
+            break;
+        case Message:
+            sv_catpvf(cxt, "%" SVf ".", it->field->name);
+            break;
+        }
+    }
+
+    SvCUR_set(cxt, SvCUR(cxt) - 1); // chop last '.'
+
+    sv_catpvs(cxt, "': ");
+    sv_catsv(cxt, warning);
+
+    if (chained_handler) {
+        dSP;
+
+        PUSHMARK(SP);
+        XPUSHs(cxt);
+        PUTBACK;
+
+        call_sv(chained_handler, G_DISCARD|G_VOID);
+    } else
+        warn_sv(cxt);
+}
